"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.getNearestFromTableFromPos = exports.getAllNestedFromNodes = exports.findColumnAtPosition = exports.createTablesFromFromNodes = exports.makeColumnName = exports.getColumnRefByPos = exports.isTableMatch = exports.getAliasFromFromTableNode = exports.createCandidatesForExpectedLiterals = exports.toCompletionItemForKeyword = exports.toCompletionItemForAlias = exports.toCompletionItemForFunction = exports.isPosInLocation = exports.makeTableName = exports.getLastToken = exports.getRidOfAfterPosString = exports.makeTableAlias = exports.ICONS = void 0;
const vscode_languageserver_types_1 = require("vscode-languageserver-types");
const log4js_1 = __importDefault(require("log4js"));
const logger = log4js_1.default.getLogger();
exports.ICONS = {
    KEYWORD: vscode_languageserver_types_1.CompletionItemKind.Text,
    COLUMN: vscode_languageserver_types_1.CompletionItemKind.Interface,
    TABLE: vscode_languageserver_types_1.CompletionItemKind.Field,
    FUNCTION: vscode_languageserver_types_1.CompletionItemKind.Property,
    ALIAS: vscode_languageserver_types_1.CompletionItemKind.Variable,
    UTILITY: vscode_languageserver_types_1.CompletionItemKind.Event,
};
const UNDESIRED_LITERAL = [
    '+',
    '-',
    '*',
    '$',
    ':',
    'COUNT',
    'AVG',
    'SUM',
    'MIN',
    'MAX',
    '`',
    '"',
    "'",
];
function isNotEmpty(value) {
    return value === null || value === undefined ? false : true;
}
function makeTableAlias(tableName) {
    if (tableName.length > 3) {
        return tableName.substring(0, 3);
    }
    return tableName;
}
exports.makeTableAlias = makeTableAlias;
function getRidOfAfterPosString(sql, pos) {
    return sql
        .split('\n')
        .filter((_v, idx) => pos.line >= idx)
        .map((v, idx) => (idx === pos.line ? v.slice(0, pos.column) : v))
        .join('\n');
}
exports.getRidOfAfterPosString = getRidOfAfterPosString;
// Gets the last token from the given string considering that tokens can contain dots.
function getLastToken(sql) {
    const match = sql.match(/^(?:.|\s)*[^A-z0-9\\.:'](.*?)$/);
    if (match) {
        let prevToken = '';
        let currentToken = match[1];
        while (currentToken != prevToken) {
            prevToken = currentToken;
            currentToken = prevToken.replace(/\[.*?\]/, '');
        }
        return currentToken;
    }
    return sql;
}
exports.getLastToken = getLastToken;
function makeTableName(table) {
    if (table.catalog) {
        return table.catalog + '.' + table.database + '.' + table.tableName;
    }
    else if (table.database) {
        return table.database + '.' + table.tableName;
    }
    return table.tableName;
}
exports.makeTableName = makeTableName;
function isPosInLocation(location, pos) {
    return (location.start.line === pos.line + 1 &&
        location.start.column <= pos.column &&
        location.end.line === pos.line + 1 &&
        location.end.column >= pos.column);
}
exports.isPosInLocation = isPosInLocation;
function toCompletionItemForFunction(f) {
    const item = {
        label: f.name,
        detail: 'function',
        kind: exports.ICONS.FUNCTION,
        documentation: f.description,
    };
    return item;
}
exports.toCompletionItemForFunction = toCompletionItemForFunction;
function toCompletionItemForAlias(alias) {
    const item = {
        label: alias,
        detail: 'alias',
        kind: exports.ICONS.ALIAS,
    };
    return item;
}
exports.toCompletionItemForAlias = toCompletionItemForAlias;
function toCompletionItemForKeyword(name) {
    const item = {
        label: name,
        kind: exports.ICONS.KEYWORD,
        detail: 'keyword',
    };
    return item;
}
exports.toCompletionItemForKeyword = toCompletionItemForKeyword;
// Check if parser expects us to terminate a single quote value or double quoted column name
// SELECT TABLE1.COLUMN1 FROM TABLE1 WHERE TABLE1.COLUMN1 = "hoge.
// We don't offer the ', the ", the ` as suggestions
function createCandidatesForExpectedLiterals(nodes) {
    const literals = nodes.map((v) => v.text);
    const uniqueLiterals = [...new Set(literals)];
    return uniqueLiterals
        .filter((v) => !UNDESIRED_LITERAL.includes(v))
        .map((v) => {
        switch (v) {
            case 'ORDER':
                return 'ORDER BY';
            case 'GROUP':
                return 'GROUP BY';
            case 'LEFT':
                return 'LEFT JOIN';
            case 'RIGHT':
                return 'RIGHT JOIN';
            case 'INNER':
                return 'INNER JOIN';
            default:
                return v;
        }
    })
        .map((v) => toCompletionItemForKeyword(v));
}
exports.createCandidatesForExpectedLiterals = createCandidatesForExpectedLiterals;
function getAliasFromFromTableNode(node) {
    if (node.as) {
        return node.as;
    }
    if (node.type === 'table') {
        return node.table;
    }
    return '';
}
exports.getAliasFromFromTableNode = getAliasFromFromTableNode;
/**
 * Test if the given table matches the fromNode.
 * @param fromNode
 * @param table
 * @returns
 */
function isTableMatch(fromNode, table) {
    switch (fromNode.type) {
        case 'subquery': {
            if (fromNode.as && fromNode.as !== table.tableName) {
                return false;
            }
            break;
        }
        case 'table': {
            if (fromNode.table && fromNode.table !== table.tableName) {
                return false;
            }
            if (fromNode.db && fromNode.db !== table.database) {
                return false;
            }
            if (fromNode.catalog && fromNode.catalog !== table.catalog) {
                return false;
            }
            break;
        }
        default: {
            return false;
        }
    }
    return true;
}
exports.isTableMatch = isTableMatch;
function getColumnRefByPos(columns, pos) {
    return columns.find((v) => 
    // guard against ColumnRefNode that don't have a location,
    // for example sql functions that are not known to the parser
    v.location &&
        v.location.start.line === pos.line + 1 &&
        v.location.start.column <= pos.column &&
        v.location.end.line === pos.line + 1 &&
        v.location.end.column >= pos.column);
}
exports.getColumnRefByPos = getColumnRefByPos;
function makeColumnName(alias, columnName) {
    return alias ? alias + '.' + columnName : columnName;
}
exports.makeColumnName = makeColumnName;
function createTablesFromFromNodes(fromNodes) {
    return fromNodes.reduce((p, c) => {
        var _a;
        if (c.type !== 'subquery') {
            return p;
        }
        if (!Array.isArray(c.subquery.columns)) {
            return p;
        }
        const columns = c.subquery.columns
            .map((v) => {
            if (typeof v === 'string') {
                return null;
            }
            return {
                columnName: v.as || (v.expr.type === 'column_ref' && v.expr.column) || '',
                description: 'alias',
            };
        })
            .filter(isNotEmpty);
        return p.concat({
            database: null,
            catalog: null,
            columns: columns !== null && columns !== void 0 ? columns : [],
            tableName: (_a = c.as) !== null && _a !== void 0 ? _a : '',
        });
    }, []);
}
exports.createTablesFromFromNodes = createTablesFromFromNodes;
function findColumnAtPosition(ast, pos) {
    var _a, _b;
    const columns = ast.columns;
    if (Array.isArray(columns)) {
        // columns in select clause
        const columnRefs = columns
            .map((col) => col.expr)
            .filter((expr) => expr.type === 'column_ref');
        if (ast.type === 'select' && ((_a = ast.where) === null || _a === void 0 ? void 0 : _a.expression)) {
            if (ast.where.expression.type === 'column_ref') {
                // columns in where clause
                columnRefs.push(ast.where.expression);
            }
        }
        // column at position
        const columnRef = getColumnRefByPos(columnRefs, pos);
        if (logger.isDebugEnabled())
            logger.debug(JSON.stringify(columnRef));
        return columnRef !== null && columnRef !== void 0 ? columnRef : null;
    }
    else if (columns.type == 'star') {
        if (ast.type === 'select' && ((_b = ast.where) === null || _b === void 0 ? void 0 : _b.expression)) {
            // columns in where clause
            const columnRefs = ast.where.expression.type === 'column_ref' ? [ast.where.expression] : [];
            // column at position
            const columnRef = getColumnRefByPos(columnRefs, pos);
            if (logger.isDebugEnabled())
                logger.debug(JSON.stringify(columnRef));
            return columnRef !== null && columnRef !== void 0 ? columnRef : null;
        }
    }
    return null;
}
exports.findColumnAtPosition = findColumnAtPosition;
/**
 * Recursively pull out the FROM nodes (including sub-queries)
 * @param tableNodes
 * @returns
 */
function getAllNestedFromNodes(tableNodes) {
    return tableNodes.flatMap((tableNode) => {
        var _a;
        let result = [tableNode];
        if (tableNode.type == 'subquery') {
            const subTableNodes = ((_a = tableNode.subquery.from) === null || _a === void 0 ? void 0 : _a.tables) || [];
            result = result.concat(getAllNestedFromNodes(subTableNodes));
        }
        return result;
    });
}
exports.getAllNestedFromNodes = getAllNestedFromNodes;
/**
 * Finds the most deeply nested FROM node that have a range encompasing the position.
 * In cases such as SELECT * FROM T1 JOIN (SELECT * FROM (SELECT * FROM T2 <pos>))
 * We will get a list of nodes like this
 * SELECT * FROM T1
 * (SELECT * FROM
 *    (SELECT * FROM T2))
 * The idea is to reverse the list so that the most nested queries come first. Then
 * apply a filter to keep only the FROM nodes which encompass the position and take
 * the first one from that resulting list.
 * @param fromNodes
 * @param pos
 * @returns
 */
function getNearestFromTableFromPos(fromNodes, pos) {
    var _a;
    return ((_a = fromNodes
        .reverse()
        .filter((tableNode) => isPosInLocation(tableNode.location, pos))
        .shift()) !== null && _a !== void 0 ? _a : null);
}
exports.getNearestFromTableFromPos = getNearestFromTableFromPos;
//# sourceMappingURL=utils.js.map